"    vim-hec5.vim
"    Copyright (C) 2022 НЕВСКИЙ БЛЯДИНА <neva_blyad@lovecry.pt>
"                                       <neva_blyad@lovecri.es>
"
"    This program is free software: you can redistribute it and/or modify
"    it under the terms of the GNU General Public License as published by
"    the Free Software Foundation, either version 3 of the License, or
"    (at your option) any later version.
"
"    This program is distributed in the hope that it will be useful,
"    but WITHOUT ANY WARRANTY; without even the implied warranty of
"    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
"    GNU General Public License for more details.
"
"    You should have received a copy of the GNU General Public License
"    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"
"  _____                                  ___             
" |  ___|__  _ __   _   _  ___  _   _    |_ _|_ __ __ _   
" | |_ / _ \| '__| | | | |/ _ \| | | |    | || '__/ _` |  
" |  _| (_) | |    | |_| | (_) | |_| |_   | || | | (_| |_ 
" |_|  \___/|_|     \__, |\___/ \__,_( ) |___|_|  \__,_( )
"                   |___/            |/                |/ 
"   __                                         _                     _   _ 
"  / _|_ __ ___  _ __ ___    _ __ ___  _   _  | |__   ___  __ _ _ __| |_| |
" | |_| '__/ _ \| '_ ` _ \  | '_ ` _ \| | | | | '_ \ / _ \/ _` | '__| __| |
" |  _| | | (_) | | | | | | | | | | | | |_| | | | | |  __/ (_| | |  | |_|_|
" |_| |_|  \___/|_| |_| |_| |_| |_| |_|\__, | |_| |_|\___|\__,_|_|   \__(_)
"                                      |___/                                
" 
" (Also for Natasha and Dasha; forever in my memory.)

" UTF-8
scriptencoding utf-8

" Do HEX-to-ASCII conversion (and vice versa) for non-latin 1-byte encodings and
" even UTF-8.
" It decreases speed dramatically. (I'll fix it in nearest futures.)
"
" Enabled by default.
if !exists('g:hec5_ascii_extended')
    let g:hec5_ascii_extended = 1
endif

" HEX bytes case.
"
" Lowercase by default.
if !exists('g:hec5_hex_uppercase')
    let g:hec5_hex_uppercase = 0
endif

" The currently edited byte
let b:col  = 0
let b:mod5 = 0
let b:len  = 0
let b:num  = 0

let b:col_hex   = 0
let b:col_ascii = 0

" Convert hex to ASCII.
"
" Each item in the parameter should have the 4-characters format, e. g. AH14.
" Item delimiter is space.
function! s:Hex2ASCII(hex)
    if b:enc == "utf-8"
        let l:cmd = "perl -e 'foreach (@ARGV)
                   \          {
                   \              my $out = pack q(H4), $_;
                   \
                   \              for (0, 1)
                   \              {
                   \                  my $ch = substr $out, $_, 1;
                   \                  print($ch ge q( ) && $ch le q(~) ? $ch : q(.));
                   \              }
                   \          }' " . a:hex
    else
        let l:cmd = "perl -e 'use Text::Iconv;
                   \
                   \          my $iconv = Text::Iconv->new(qw(" . b:enc . ", utf-8));
                   \
                   \          foreach (@ARGV)
                   \          {
                   \              for my $idx (0, 2)
                   \              {
                   \                  my $hex = substr $_, $idx, 2;
                   \                  my $dec = hex $hex;
                   \                  my $ch  = pack q(S), $dec;
                   \
                   \                  my $conv_ch = $iconv->convert($ch);
                   \                  $conv_ch = substr $conv_ch, 0, -1;
                   \                  print($conv_ch =~ m/\\p{Print}/ ? $conv_ch : q(.));
                   \                  
                   \                  last if length == 2;
                   \              }
                   \          }' " . a:hex
    endif

    let l:ascii = system(l:cmd)
    return l:ascii
endfunction

" Convert to ASCII hex.
"
" Each item in the returned value should have the 4-characters format, e. g.
" AH14.
" Item delimiter is space.
function! s:ASCII2Hex(ascii)
    let l:ascii = escape(a:ascii, "'") 

    if b:enc == "utf-8"
        let l:cmd = "perl -e 'use Encode;
                   \
                   \          Encode::_utf8_on($ARGV[0]);
                   \          my $len = length $ARGV[0];
                   \
                   \          for (my $idx = 0; $idx < $len; $idx += 2)
                   \          {
                   \              my $in  = substr $ARGV[0], $idx, 2;
                   \              my $hex = unpack 'H4', $in;
                   \
                   \              print uc $hex, q( );
                   \          }' '" . l:ascii . "'"
    else
        let l:cmd = "perl -e 'use Encode;
                   \          use Text::Iconv;
                   \
                   \          my $iconv = Text::Iconv->new(qw(utf-8, " . b:enc . "));
                   \          Encode::_utf8_on($ARGV[0]);
                   \          my $len = length $ARGV[0];
                   \
                   \          for (my $idx = 0; $idx < $len; $idx++)
                   \          {
                   \              my $ch      = substr $ARGV[0], $idx, 1;
                   \              my $conv_ch = $iconv->convert($ch);
                   \              my $dec     = unpack 'C', $conv_ch;
                   \              my $hex     = sprintf q(%X), $dec;
                   \
                   \              print $hex, $idx % 2 == 0 ? q() : q( );
                   \          }' '" . l:ascii . "'"
    endif

    let l:hex = system(l:cmd)
    return l:hex
endfunction

" Highlight byte under the cursor in HEX block and the corresponding character
" in ASCII block.
"
" Do vice versa when editing in ASCII block.
function! s:CurHighlight()
    let pos        = getcurpos()
    let set_syntax = v:false

    let l:line   = pos[1]
    let b:col    = pos[4]
    let max_line = line("$")
    let b:mod5   = b:col % 5

    if l:line == max_line
        let l:line_str = getline(l:line)
        let l:pattern  = '^\x\{8}: \(\%(\x\{4} \)\{-}\%(\x\{2}\)\{1,2}\) \{-1,}.\{1,16}$'
        let l:matches  = matchlist(l:line_str, l:pattern)

        if l:matches == []
            undo
            return
        endif

        let b:len      = len(l:matches[1])
        let b:num      = (b:len - (b:len / 4 - 1)) 
    else
        let b:len      = 39
        let b:num      = 16
    endif

    if b:col >= 11 && b:col <= 11 + b:len - 1 && b:mod5 != 0
        " We are in the HEX block.
        "
        " Find character in ASCII block.
        let set_syntax  = v:true
        let b:col_hex   = b:col - (b:mod5 == 2 || b:mod5 == 4)
        let b:col_ascii = 2 * ((b:col_hex - 11) / 5) + 52 + (b:mod5 == 3 || b:mod5 == 4)
    elseif b:col >= 52 && b:col <= 52 + b:num - 1
        " We are in the ASCII block.
        "
        " Find byte in HEX block.
        let set_syntax  = v:true
        let b:col_ascii = b:col
        let b:col_hex   = (b:col_ascii - 52) / 2 * 5 + 11 + 2 * (b:col_ascii % 2)
    endif

    syntax clear cur
    syntax clear xxdAscii
    if set_syntax
        let l:pattern = '"\%' . l:line . 'l\%(\%' . b:col_hex . 'c\x\x\|\%' . b:col_ascii . 'v.\)"'
        execute 'syntax match cur ' . l:pattern
        highlight default link cur Identifier
    endif
    syntax match xxdAscii "  .\{,16\}\r\=$"hs=s+2 contains=xxdDot,cur
endfunction

" Update ASCII representation of the currently selected byte
function! s:LineUpd(line, restore_pos)
    let l:line_str = getline(a:line)
    let l:pattern  = '^\(\x\{8}\): \(\%(\x\{4} \)\{-}\%(\x\{2}\)\{1,2}\) \{-1,}\(.\{1,16}\)$'
    let l:matches  = matchlist(l:line_str, l:pattern)

    if l:matches == []
        undo
        return
    endif

    let l:pos = getpos(".")

    if l:pos[2] <= 51
        " We are in the HEX block.
        "
        " Convert it to ASCII.
        let l:spaces   = repeat(" ", 40 - len(l:matches[2]))
        let l:ascii    = s:Hex2ASCII(l:matches[2])
        let l:str      = l:matches[1] . ": " . l:matches[2] . " " . l:spaces . l:ascii
    else
        " We are in the ASCII block.
        "
        " Convert it to HEX.
        let l:pattern  = '\.'
        let l:matches2 = matchlist(l:matches[3], l:pattern)

        if l:matches2 != []
            undo
            return
        endif

        let l:hex      = s:ASCII2Hex(l:matches[3])
        let l:spaces   = repeat(" ", 40 - len(l:hex))
        let l:str      = l:matches[1] . ": " . l:hex . " " . l:spaces . l:matches[3]
    endif

    call setline(a:line, l:str)

    if a:restore_pos
        call setpos(".", l:pos)
    endif
endfunction

" Cycle between HEX and ASCII blocks
function! <SID>BlocksCycle()
    if b:col >= 11 && b:col <= 11 + b:len - 1 && b:mod5 != 0
        " We are in the HEX block.
        "
        " Go to ASCII.
        execute 'normal ' . b:col_ascii . '|'
    elseif b:col >= 52 && b:col <= 52 + b:num - 1
        " We are in the ASCII block.
        "
        " Go to HEX.
        execute 'normal ' . b:col_hex . '|'
    endif
endfunction

" Edit binary using hex editor
command! HexConvert call <SID>HexConvert()

function! <SID>HexConvert()
    " Check filetype
    if &filetype == "xxd"
        " We're already editing xxd file.
        " Convert back.
        if b:bin
            setlocal filetype=
            setlocal fenc=
            setlocal binary
            setlocal endofline
            setlocal nofixendofline
        else
            setlocal filetype=
            setlocal nobinary
            setlocal endofline
            setlocal fixendofline
        endif

        %!xxd -r
    else
        " We're editing not-xxd file.
        " Do conversion.
        let b:enc = &fenc
        let b:bin = &binary

        if g:hec5_hex_uppercase
            %!xxd -u
        else
            %!xxd
        endif

        setlocal filetype=xxd

        " Do conversion to ASCII
        if g:hec5_ascii_extended
            let line     = 1
            let max_line = line("$")

            while line <= max_line
                call s:LineUpd(line, v:false)
                let line += 1
            endwhile
            call setpos(".", [ 0, 1, 11, 0 ])
        endif

        syntax match cur "$"

        " Add mappings, unless the user didn't want this
        if !exists("no_plugin_maps")
            nnoremap <Plug>hexblockscycle :call <SID>BlocksCycle()<CR>
        endif

        " Add autocommands
        autocmd CursorMoved,CursorMovedI <buffer> :call s:CurHighlight()
        if g:hec5_ascii_extended
            autocmd TextChanged,TextChangedI <buffer> :call s:LineUpd(line("."), v:true)
        endif
    endif
endfunction
